use modules::EventLoop;

use toml::Value;
use toml::value::Table;

use hlua;
use hlua::{Lua, LuaFunction, AnyHashableLuaValue, AnyLuaValue};

use {User, Message, Channel, NullMessageSender};
use event::{Event, Envelope};

use std::sync::Arc;
use std::sync::mpsc::Receiver;
use transformable_channels::mpsc::ExtSender;

use std::path::Path;
use std::fs::File;

use std::collections::HashMap;

pub struct LuaModule {
    code: Option<String>,
    file: Option<String>,
    variables: Table
}

impl LuaModule {
    pub fn new (_: &Table, config: &Table) -> Box<EventLoop> {
        Box::new(LuaModule {
            code: config.get("code").and_then(|value| value.as_str()).map(String::from),
            file: config.get("file").and_then(|value| value.as_str()).map(String::from),
            variables: config.clone()
        })
    }
}

fn set (lua: &mut Lua, key: &str, value: &Value) {
    match value.clone() {
        Value::String(string) => lua.set(key, string),
        Value::Integer(integer) => lua.set(key, integer as i32),
        Value::Float(float) => lua.set(key, float),
        Value::Boolean(boolean) => lua.set(key, boolean),
        _ => {}
    }
}

struct MessageWrapper {
    envelope: Arc<Envelope>
}

implement_lua_read!(MessageWrapper);
implement_lua_push!(MessageWrapper, |mut metatable| {
    let mut index = metatable.empty_array("__index");
    index.set("content", hlua::function1(|wrapper: &MessageWrapper| {
        if let Event::Message { ref message } = wrapper.envelope.event {
            Some(message.content.clone())
        } else {
            None
        }
    }));
    index.set("reply", hlua::function2(|wrapper: &MessageWrapper, reply: String| {
        if let Event::Message { ref message } = wrapper.envelope.event {
            message.reply(&reply);
        }
    }));
});

struct SenderWrapper {
    sender: Box<ExtSender<Event>>
}

implement_lua_read!(SenderWrapper);
implement_lua_push!(SenderWrapper, |mut metatable| {
    let mut index = metatable.empty_array("__index");
    index.set("send", hlua::function2(|wrapper: &SenderWrapper, data: HashMap<AnyHashableLuaValue, AnyLuaValue>| {
        if let Some(&AnyLuaValue::LuaString(ref event_type)) = data.get(&AnyHashableLuaValue::LuaString("type".to_owned())) {
            match event_type.as_ref() {
                "message" => {
                    wrapper.sender.send(Event::Message {
                        message: Message {
                            author: User {
                                name: data.get(&AnyHashableLuaValue::LuaString("username".to_owned()))
                                             .and_then(|value| match value {
                                                 &AnyLuaValue::LuaString (ref string_value) => Some(string_value.to_owned()),
                                                 _ => None
                                             })
                                             .unwrap_or_else(|| "".to_owned()),
                                sender: Box::new(NullMessageSender {})
                            },
                            channel: data.get(&AnyHashableLuaValue::LuaString("channel".to_owned()))
                                         .and_then(|value| match value {
                                             &AnyLuaValue::LuaString (ref string_value) => Some(string_value.to_owned()),
                                             _ => None
                                         }).map(|channel| Channel {
                                             name: channel,
                                             description: "".to_owned(),
                                             sender: Box::new(NullMessageSender {}),
                                             topic: "".to_owned()
                                         }),
                            content: data.get(&AnyHashableLuaValue::LuaString("message".to_owned()))
                                         .and_then(|value| match value {
                                             &AnyLuaValue::LuaString (ref string_value) => Some(string_value.to_owned()),
                                             _ => None
                                         })
                                         .expect("Invalid message event passed in")
                        }
                    });
                },
                _ => {}
            }
        }
    }));
});

impl EventLoop for LuaModule {
    fn run (&self, sender: Box<ExtSender<Event>>, receiver: Receiver<Arc<Envelope>>) {
        let mut lua = Lua::new();
        lua.openlibs();

        for (key, value) in &self.variables {
            set(&mut lua, key, value);
        }

        lua.set("sender", SenderWrapper {
            sender: Box::new(sender.clone())
        });

        match self.code {
            Some(ref code) => { lua.execute(&code).unwrap() },
            None => {
                match self.file {
                    Some(ref file) => { lua.execute_from_reader::<(), _>(File::open(&Path::new(file)).unwrap()).unwrap(); }
                    None => {}
                }
            }
        }

        loop {
            match receiver.recv() {
                Ok(envelope) => {
                    match envelope.event {
                        Event::Configure { ref configuration } => {
                            for (key, value) in configuration {
                                set(&mut lua, key, value);
                            }
                        },
                        Event::Message { ref message } => {
                            let on_message: Option<LuaFunction<_>> = lua.get("on_message");
                            match on_message {
                                Some(mut on_message) => {
                                    on_message.call_with_args::<(), _, _>(MessageWrapper {
                                        envelope: envelope.clone()
                                    }).unwrap();
                                },
                                None => {}
                            }
                        }
                        _ => ()
                    }
                }
                Err(error) => { error!("Error {:?}", error) }
            }
        }
    }
}
